// joystick.cc -- Class object for the joystick device

// Standard function header files
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

// Standard X11 header files
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>

// X Input Extention header files
#include <X11/extensions/XInput.h>

// Private include files
#include "joystick.h"


// Initialize the Joystick class object
Joystick::Joystick (void) {
  display = NULL;
  device = NULL;
}


// Release the joystick
Joystick::~Joystick () {
  if (device) {
  }
}


// Open the joystick device
int Joystick::initialize (Display *disp) {
  int num_devices, i, t2, t3;
  XDeviceInfo *xdevices;
  XAnyClassPtr any;
  XValuatorInfo *stick;

  display = disp;
  if (!XQueryExtension (display, "XInputExtension",
			&num_devices, &t2, &t3)) {
    cerr << "XInputExtension not found; joystick not available\n";
    return 0;
  }

  xdevices = XListInputDevices (display, &num_devices);
  for (i = 0; i < num_devices; i++) {
    if (xdevices[i].use != IsXExtensionDevice) continue;
    if (strcmp (xdevices[i].name, "Joystick") != 0) continue;
    device = XOpenDevice (display, xdevices[i].id);
    if (device) break;
  }
  XFreeDeviceList (xdevices);

  if (!device) {
    cerr << "Joystick not found among XInputExtension devices\n";
    return 0;
  }

  stick = (XValuatorInfo *) xdevices[i].inputclassinfo;
  while (stick->c_class != ValuatorClass) {
    stick = (XValuatorInfo *) ((char *) stick + stick->length);
  }
  x_min = stick->axes[0].min_value;
  x_max = stick->axes[0].max_value;
  xc_min = (2 * x_min + x_max) / 3;
  xc_max = (x_min + 2 * x_max) / 3;
  y_min = stick->axes[1].min_value;
  y_max = stick->axes[1].max_value;
  yc_min = (2 * y_min + y_max) / 3;
  yc_max = (y_min + 2 * y_max) / 3;
  calibrate ();
}


// Get the current relative x direction
int Joystick::x_direction (void) {
  XValuatorState *stick;

  get_stick (&stick);
  if (stick->valuators[0] < xc_min)
    return -1;
  if (stick->valuators[0] > xc_max)
    return 1;
  return 0;
}


// Get the current absolute x position
double Joystick::x_position (void) {
  XValuatorState *stick;

  get_stick (&stick);
  if (stick->valuators[0] < xc_min)
    return ((float) (stick->valuators[0] - xc_min) / (float) (xc_min - x_min));
  if (stick->valuators[0] > xc_max)
    return ((float) (stick->valuators[0] - xc_max) / (float) (x_max - xc_max));
  return 0.0;
}


// Get the current relative y direction
int Joystick::y_direction (void) {
  XValuatorState *stick;

  get_stick (&stick);
  if (stick->valuators[1] < yc_min)
    return -1;
  if (stick->valuators[1] > yc_max)
    return 1;
  return 0;
}


// Get the current absolute y position
double Joystick::y_position (void) {
  XValuatorState *stick;

  get_stick (&stick);
  if (stick->valuators[1] < yc_min)
    return ((float) (stick->valuators[1] - yc_min) / (float) (yc_min - y_min));
  if (stick->valuators[1] > yc_max)
    return ((float) (stick->valuators[1] - yc_max) / (float) (y_max - yc_max));
  return 0.0;
}


// Get the current button state
int Joystick::buttons (void) {
  XDeviceState *state;
  XButtonState *buttons;
  int i;

  state = XQueryDeviceState (display, device);
  buttons = (XButtonState *) state->data;
  while (buttons->c_class != ButtonClass) {
    buttons = (XButtonState *) ((char *) buttons + buttons->length);
  }

  return buttons->buttons[0] >> 1;
}


// Get the raw joystick data
void Joystick::get_stick (XValuatorState **stick) {
  XDeviceState *state;

  state = XQueryDeviceState (display, device);
  *stick = (XValuatorState *) state->data;
  while ((*stick)->c_class != ValuatorClass) {
    (*stick) = (XValuatorState *) ((char *) (*stick) + (*stick)->length);
  }
  return;
}


// Calibrate the joystick
void Joystick::calibrate (void) {
  Window win;
  XWMHints *wm_hints;
  unsigned int wx, wy, ww, wh, dump;
  XSizeHints *size_hints;
  char *namestring = "Joystick Calibration";
  XTextProperty windowName;
  XFontStruct *font_info;
  unsigned int tw, th;
  char *current_message = "Please move the joystick in full circles";
  GC gc;
  XEvent event;
  XDeviceState *devstate;
  XValuatorState *stickstate;
  XButtonState *buttonstate;
  enum { CIRCLING, TOPLEFT, CENTERMIN,
	 BOTTOMRIGHT, CENTERMAX, FINISHED } cal_state = CIRCLING;
  enum { ATMIN, ATMAX } circle_x_state = ATMIN, circle_y_state = ATMIN;
  int circle_x_count = 0, circle_y_count = 0;
  struct timeval timev;
  struct timezone timez = { 0, 0 };
  unsigned int old_time, new_time;
  fstream *cal_data;
  char *cal_fname;
  const char *home_dir;

  // See of a calibration file exists, and if so, use it.
  home_dir = getenv ("HOME");
  if (home_dir == NULL)
    home_dir = ".";
  cal_fname = new char [strlen (home_dir)
		       + strlen ("/.joystick_calibration") + 1];
  if (cal_fname != NULL)
    {
      strcpy (cal_fname, home_dir);
      strcat (cal_fname, "/.joystick_calibration");
      cal_data = new fstream (cal_fname, ios::in);
      if (*cal_data)
	{
	  *cal_data >> x_min >> xc_min >> xc_max >> x_max >>
	    y_min >> yc_min >> yc_max >> y_max;
	  if (*cal_data)
	    {
	      // If everything read ok, return without calibrating again.
	      delete cal_data;
	      return;
	    }
	}
      delete cal_data;
    }

  // Create a window for joystick feedback.
  XGetGeometry (display, DefaultRootWindow (display), &win,
		(int *) &dump, (int *) &dump, &wx, &wy, &dump, &dump);
  if ((font_info = XLoadQueryFont (display, "fixed"))) {
    th = font_info->ascent + font_info->descent;
    wh = (th + 2) * 2;
    tw = XTextWidth (font_info, current_message, strlen (current_message));
    ww = tw + th + 2;
  } else {
    ww = wh = 100;
  }
  wx = (wx - ww) / 2;
  wy = (wy - wh) / 2;
  win = XCreateSimpleWindow (display, DefaultRootWindow (display),
			     wx, wy, ww, wh, 0,
			     BlackPixel (display, DefaultScreen (display)),
			     WhitePixel (display, DefaultScreen (display)));
  if (XStringListToTextProperty (&namestring, 1, &windowName)) {
    XSetWMName (display, win, &windowName);
    XFree (windowName.value);
  }
  if ((wm_hints = XAllocWMHints ())) {
    wm_hints->initial_state = NormalState;
    wm_hints->input = True;
    wm_hints->flags = StateHint | InputHint;
    XSetWMHints (display, win, wm_hints);
    XFree (wm_hints);
  }
  if ((size_hints = XAllocSizeHints ())) {
    size_hints->flags = PPosition | PSize | PMinSize | PMaxSize | PWinGravity;
    size_hints->min_width = size_hints->max_width = 100;
    size_hints->min_height = size_hints->max_height = 100;
    size_hints->win_gravity = CenterGravity;
    XSetWMNormalHints (display, win, size_hints);
    XFree (size_hints);
  }
  XSelectInput (display, win, ExposureMask);
  gc = XCreateGC (display, win, 0, NULL);
  XSetForeground (display, gc, BlackPixel (display, DefaultScreen (display)));
  XSetBackground (display, gc, WhitePixel (display, DefaultScreen (display)));
  if (font_info)
    XSetFont (display, gc, font_info->fid);
  XMapWindow (display, win);

  // Now run our own event loop, displaying the current message as needed,
  // grabbing joystick events, and updating our ranges.
  x_min = x_max = 0;
  xc_min = xc_max = 0;
  y_min = y_max = 0;
  yc_min = yc_max = 0;
  while (cal_state != FINISHED) {
    if (XPending (display)) {
      // Very small routine to display the current message
      XNextEvent (display, &event);
      switch (event.type) {
      case Expose:
	if (!event.xexpose.count) {
	  XClearArea (display, win, 0, 0, 0, 0, False);
	  XDrawImageString (display, win, gc,
			    (ww - tw) / 2, (wh - th) / 2 + font_info->ascent,
			    current_message, strlen (current_message));
	}
	break;
      }
    } else {
      // See where the joystick is pointed now
      get_stick (&stickstate);
      switch (cal_state) {
      case CIRCLING:
	// When we reach any of the four sides, reset our limits.
	// If we have just come from the opposite side,
	//  increment our travel count.
	if (stickstate->valuators[0] <= x_min) {
	  x_min = stickstate->valuators[0];
	  if (circle_x_state == ATMAX) {
	    circle_x_state = ATMIN;
	    circle_x_count++;
	  }
	} else if (stickstate->valuators[0] >= x_max) {
	  x_max = stickstate->valuators[0];
	  if (circle_x_state == ATMIN) {
	    circle_x_state = ATMAX;
	    circle_x_count++;
	  }
	}
	if (stickstate->valuators[1] <= y_min) {
	  y_min = stickstate->valuators[1];
	  if (circle_y_state == ATMAX) {
	    circle_y_state = ATMIN;
	    circle_y_count++;
	  }
	} else if (stickstate->valuators[1] >= y_max) {
	  y_max = stickstate->valuators[1];
	  if (circle_y_state == ATMIN) {
	    circle_y_state = ATMAX;
	    circle_y_count++;
	  }
	}
	// It takes at least three full circles to ensure we went everywhere.
	if ((circle_x_count >= 6) && (circle_y_count >= 6)) {
	  cal_state = TOPLEFT;
	  current_message = "Now move the joystick forward and left";
	  tw = XTextWidth (font_info, current_message,
			   strlen (current_message));
	  XClearArea (display, win, 0, 0, 0, 0, True);
	  gettimeofday (&timev, &timez);
	  old_time = new_time = timev.tv_sec;
	}
	break;
      case TOPLEFT:
	// Pause with the new message before accepting any new values.
	if (new_time - old_time < 3) {
	  gettimeofday (&timev, &timez);
	  new_time = timev.tv_sec;
	  break;
	}
	if ((stickstate->valuators[0] < (3 * x_min + x_max) / 4) &&
	    (stickstate->valuators[1] < (3 * y_min + y_max) / 4)) {
	  cal_state = CENTERMIN;
	  current_message = "Now move the joystick back to center";
	  tw = XTextWidth (font_info, current_message,
			   strlen (current_message));
	  XClearArea (display, win, 0, 0, 0, 0, True);
	  gettimeofday (&timev, &timez);
	  old_time = new_time = timev.tv_sec;
	}
	break;
      case CENTERMIN:
	gettimeofday (&timev, &timez);
	if ((stickstate->valuators[0] > (x_min + 2 * x_max) / 3) ||
	    (stickstate->valuators[1] > (y_min + 2 * y_max) / 3)) {
	  // They went the wrong way; go back a step.
	  cal_state = TOPLEFT;
	  current_message = "Now move the joystick forward and left";
	  tw = XTextWidth (font_info, current_message,
			   strlen (current_message));
	  XClearArea (display, win, 0, 0, 0, 0, True);
	  old_time = new_time = timev.tv_sec;
	// Wait for them to get within the middle third.
	} else if ((stickstate->valuators[0] < (2 * x_min + x_max) / 3) ||
		   (stickstate->valuators[1] < (2 * y_min + y_max) / 3)) {
	  old_time = timev.tv_sec;
	// Spend at least three seconds in the middle before assuming
	//  that we're centered.
	} else if (timev.tv_sec - old_time >= 3) {
	  xc_min = stickstate->valuators[0] - 1;
	  yc_min = stickstate->valuators[1] - 1;
	  cal_state = BOTTOMRIGHT;
	  current_message = "Now move the joystick backward and right";
	  tw = XTextWidth (font_info, current_message,
			   strlen (current_message));
	  XClearArea (display, win, 0, 0, 0, 0, True);
	  old_time = new_time = timev.tv_sec;
	}
	break;
      case BOTTOMRIGHT:
	// Pause with the new message before accepting any new values.
	if (new_time - old_time < 3) {
	  gettimeofday (&timev, &timez);
	  new_time = timev.tv_sec;
	  break;
	}
	if ((stickstate->valuators[0] > (x_min + 3 * x_max) / 4) &&
	    (stickstate->valuators[1] > (y_min + 3 * y_max) / 4)) {
	  cal_state = CENTERMAX;
	  current_message = "Now move the joystick back to center";
	  tw = XTextWidth (font_info, current_message,
			   strlen (current_message));
	  XClearArea (display, win, 0, 0, 0, 0, True);
	  gettimeofday (&timev, &timez);
	  old_time = new_time = timev.tv_sec;
	}
	break;
      case CENTERMAX:
	gettimeofday (&timev, &timez);
	if ((stickstate->valuators[0] < (2 * x_min + x_max) / 3) ||
	    (stickstate->valuators[1] < (2 * y_min + y_max) / 3)) {
	  // They went the wrong way; go back a step.
	  cal_state = BOTTOMRIGHT;
	  current_message = "Now move the joystick backward and right";
	  tw = XTextWidth (font_info, current_message,
			   strlen (current_message));
	  XClearArea (display, win, 0, 0, 0, 0, True);
	  old_time = new_time = timev.tv_sec;
	// Wait for them to get within the middle third.
	} else if ((stickstate->valuators[0] > (x_min + 2 * x_max) / 3) ||
		   (stickstate->valuators[1] > (y_min + 2 * y_max) / 3)) {
	  old_time = timev.tv_sec;
	// Spend at least three seconds in the middle before assuming
	//  that we're centered.
	} else if (timev.tv_sec - old_time >= 3) {
	  xc_max = stickstate->valuators[0] + 1;
	  yc_max = stickstate->valuators[1] + 1;
	  cal_state = FINISHED;
	}
	break;
      }
    }
  }

  // Save the calibration data in a file.
  if (cal_fname != NULL)
    {
      cal_data = new fstream (cal_fname, ios::out);
      if (*cal_data)
	{
	  *cal_data << x_min << "\n" << xc_min << "\n" << xc_max << "\n" <<
	    x_max << "\n" << y_min << "\n" << yc_min << "\n" << yc_max << "\n" <<
	    y_max << "\n";
	} else {
	  cerr << "Error opening calibration file " << cal_fname << "\n";
	}
      delete cal_data;
    }

  // Free the temporary window and other data.
  XUnmapWindow (display, win);
  if (font_info)
    XFreeFont (display, font_info);
  XFreeGC (display, gc);
  return;
}
