// Interface between the Adventure and X Windows


// Public include files
#include <iostream.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


// Private include files
#include "data.h"
#include "transient.h"
#include "xgui.h"


// Data structures
static XGUI *adventure_window;
static Atom wm_protocols, wm_delete_window, xa_signal;


// Signal handler for the timeout alarm
static void
handle_signal (int sig)
{
  adventure_window->signal (sig);
}


// Class methods for the X interface
XGUI::XGUI (void)
{
  XWMHints *wm_hints;
  XClassHint *class_hints;
  XSizeHints *size_hints;
  XTextProperty windowName;
  XGCValues gcv;
  const char *windowNameString = "An Adventure Game";
  struct sigaction alrm_handler;
  int i;

  // Attempt to open the X display
  if ((display = XOpenDisplay (NULL)) == NULL)
    {
      cerr << "xANALOGadventure: cannot connect to X server "
	   << XDisplayName (NULL) << "\n";
      exit (-1);
    }

  // Do this for debugging
  // XSynchronize (display, True);

  // Figure out what font we're going to use.
  font = XLoadQueryFont (display,
			 "-adobe-utopia-medium-r-*-*-18-*-*-*-p-*-iso8859-1");
  if (!font) {
    // Our preferred font doesn't exist; revert to a fixed font.
    font = XLoadQueryFont (display, "fixed");
    if (!font) {
      cerr << "xANALOGadventure: failed to allocate a font.\n";
      exit (-1);
    }
  }

  // Attempt to intialize the joystick
  joystick.initialize (display);

  // Allocate memory for hint structures
  if (!(size_hints = XAllocSizeHints ())
      || !(wm_hints = XAllocWMHints ())
      || !(class_hints = XAllocClassHint()))
    {
      cerr << "xANALOGadventure: failure allocating memory\n";
      exit (-1);
    }

  // Initialize the hints
  size_hints->flags = PSize | PMinSize | PMaxSize | PWinGravity;
  size_hints->min_width = size_hints->max_width = MAP_WIDTH;
  status_height = 8 + NUM_STATUS_LINES * (font->ascent + font->descent);
  size_hints->min_height = size_hints->max_height = MAP_HEIGHT + status_height;
  size_hints->win_gravity = CenterGravity;

  // Create the main window
  window = XCreateSimpleWindow (display, DefaultRootWindow (display),
				0, 0, MAP_WIDTH, size_hints->min_height, 0,
				BlackPixel (display, DefaultScreen (display)),
				WhitePixel (display, DefaultScreen (display)));

  // Set window manager hints
  if (!XStringListToTextProperty ((char **) &windowNameString, 1, &windowName))
    {
      cerr << "xANALOGadventure: structure allocation for windowName failed.\n";
      exit (-1);
    }
  wm_hints->initial_state = NormalState;
  wm_hints->input = True;
  wm_hints->flags = StateHint | InputHint;
  class_hints->res_name = NULL;
  class_hints->res_class = "xANALOGadventure";
  XSetWMProperties (display, window, &windowName, &windowName, NULL, 0,
		    size_hints, wm_hints, class_hints);
  XFree (windowName.value);
  XFree (size_hints);
  XFree (wm_hints);
  XFree (class_hints);

  // Get the atom needed to delete the window
  wm_protocols = XInternAtom (display, "WM_PROTOCOLS", False);
  wm_delete_window = XInternAtom (display, "WM_DELETE_WINDOW", False);

  // Tell the window manager we want to receive delete requests
  XSetWMProtocols (display, window, &wm_delete_window, 1);

  // Get the atom needed to receive timeout events
  xa_signal = XInternAtom (display, "xA_Signal", False);

  // Select the event types we will use
  XSelectInput (display, window, ExposureMask |
		StructureNotifyMask | KeyPressMask);

  // Set the values we will need for the graphics context.
  gcv.background = WhitePixel (display, DefaultScreen (display));
  gcv.foreground = BlackPixel (display, DefaultScreen (display));
  gcv.font = font->fid;

  // Create our graphics context.
  gc = XCreateGC (display, window, GCForeground | GCBackground | GCFont, &gcv);

  // Create the map window
  map_window = new MapWindow (display, window);

  // Create the transient message window
  transient = new TransientMessages (display, window);

  // Set up our signal handler for SIGALRM
  alrm_handler.sa_handler = handle_signal;
  alrm_handler.sa_flags = 0;
  sigemptyset (&alrm_handler.sa_mask);
  sigaction (SIGALRM, &alrm_handler, &old_handler);

  // Clear the status lines
  num_status_lines = 0;
  for (i = 0; i < 3; i++)
    {
      status_line[i] = NULL;
      status_len[i] = 0;
      line_width[i] = 0;
    }
}

XGUI::~XGUI ()
{
  // Disable any pending alarms and remove our SIGALRM handler
  alarm (0);
  sigaction (SIGALRM, &old_handler, NULL);

  // Hide our windows
  hide ();

  // Destroy the transient message window
  delete transient;

  // Destroy the map window
  delete map_window;

  // Release the main window
  XDestroyWindow (display, window);

  // Close the display
  XCloseDisplay (display);
}

void
XGUI::show (void)
{
  map_window->show ();
  XMapWindow (display, window);
}

void
XGUI::hide (void)
{
  XUnmapWindow (display, window);
}

void
XGUI::redraw (void)
{
  map_window->redraw ();
  transient->redraw ();
}

void
XGUI::redraw_area (Window W, int x, int y, int w, int h)
{
  int lineheight, i;

  if (W == window)
    {
      // Redraw our own status line.
      lineheight = font->ascent + font->descent;
      for (i = 0; i < num_status_lines; i++)
	XDrawImageString (display, window, gc, (MAP_WIDTH - line_width[i]) / 2,
			  (status_height - num_status_lines * lineheight) / 2
			  + MAP_HEIGHT + i * lineheight + font->ascent,
			  status_line[i], status_len[i]);
    } else {
      // Redraw the map window and/or the popup message window.
      map_window->redraw_area (W, x, y, w, h);
      transient->redraw ();
    }
}

// This routine handles common events internally (redraw), while
// passing interaction events back to the caller (mouse, key).
XEvent *
XGUI::next_event (void)
{
  static XEvent event;
  KeySym keysym;
  char buffer[4];

  while (1)
    {
      XNextEvent (display, &event);
      switch (event.type)
	{
	  // The following events are returned to the caller:
	case ButtonPress:
	case ButtonRelease:
	case KeyPress:
	  return &event;

	case Expose:
	  // Redraw the exposed windows.
	  redraw_area (event.xexpose.window, event.xexpose.x, event.xexpose.y,
		       event.xexpose.width, event.xexpose.height);
	  break;

	case NoExpose:
	  // Nothing needs doing.
	  break;

	case ClientMessage:
	  if ((event.xclient.message_type == wm_protocols)
	      && (event.xclient.data.l[0] == wm_delete_window))
	    {
	      // The window manager told us that we should be deleted.
	      game_over = True;
	      return &event;
	    }
	  break;
	}
    }
}

// This routine handles common events internally (redraw),
// until a specified number of milliseconds have passed.
void
XGUI::event_pause (int timeout)
{
  XEvent event;
  KeySym keysym;
  char buffer[4];

  // Set an alarm timeout.
#ifdef HAVE_UALARM
  ualarm (timeout * 1000, 0);
#else
  alarm ((timeout + 999) / 1000);
#endif

  while (1)
    {
      XNextEvent (display, &event);
      switch (event.type)
	{
	  // We only process expose events in this loop.
	case Expose:
	  redraw_area (event.xexpose.window, event.xexpose.x, event.xexpose.y,
		       event.xexpose.width, event.xexpose.height);
	  break;

	case ClientMessage:
	  if ((event.xclient.message_type == wm_protocols)
	      && (event.xclient.data.l[0] == wm_delete_window))
	    {
	      // The window manager told us that we should be deleted.
	      game_over = True;
	      // Return now.
	      return;
	    }
	  else if ((event.xclient.send_event)
		   && (event.xclient.message_type == xa_signal)
		   && (event.xclient.data.l[0] == SIGALRM))
	    {
	      // Timeout.  Return now.
	      return;
	    }
	  break;
	}
    }
}

void
XGUI::signal (int sig)
{
  XClientMessageEvent event;

  event.type = ClientMessage;
  event.serial = 0;		// Am I supposed to set this to something?
  event.send_event = True;
  event.display = display;	// These aren't used by us, but
  event.window = window;	//  may be checked by the server.
  event.message_type = xa_signal;
  event.format = 32;
  event.data.l[0] = sig;
  XSendEvent (display, window, False, 0, (XEvent *) &event);
  XFlush (display);
}

void
XGUI::noise (void)
{
  XBell (display, 100);
}

void
XGUI::print (const char *line)
{
  int i;

  // Make room for the new line.
  if (num_status_lines == NUM_STATUS_LINES)
    {
      delete status_line[0];
      for (i = 0; i < num_status_lines - 1; i++)
	{
	  status_line[i] = status_line[i + 1];
	  status_len[i] = status_len[i + 1];
	  line_width[i] = line_width[i + 1];
	}
      --num_status_lines;
    }

  // Allocate storage for the line and copy it.
  status_len[num_status_lines] = strlen (line);
  status_line[num_status_lines] = new char [status_len[num_status_lines] + 1];
  strcpy (status_line[num_status_lines], line);
  line_width[num_status_lines]
    = XTextWidth (font, status_line[num_status_lines],
		  status_len[num_status_lines]);
  num_status_lines++;

  // Redraw the area.
  XClearArea (display, window, 0, MAP_HEIGHT, 0, 0, True);
}

void
XGUI::clear_status (void)
{
  int i;

  for (i = 0; i < num_status_lines; i++)
    delete status_line[i];
  num_status_lines = 0;
}


// init_x()
// Initialize the windows
void
init_x (void)
{
  adventure_window = new XGUI;
}


// show_window()
// Map the windows
void
show_window (void)
{
  adventure_window->show ();
}


// close_x()
// Release the windows
void
close_x (void)
{
  delete adventure_window;
}


// move_event_loop()
// Process X events until movement is detected
void
move_event_loop (int *dx, int *dy)
{
  XEvent *event;
  KeySym keysym;
  char buffer[4];

  *dx = *dy = 0;
  while (!*dx && !*dy && !game_over)
    {
      event = adventure_window->next_event ();
      if (event->type == KeyPress)
	{
	  XLookupString (&event->xkey, buffer, sizeof (buffer), &keysym, NULL);
	  switch (keysym)
	    {
	    case XK_Up:
	    case XK_KP_Up:
	    case XK_KP_8:
	    case XK_N:
	    case XK_n: *dy = -1; break;
	    case XK_Down:
	    case XK_KP_Down:
	    case XK_KP_2:
	    case XK_S:
	    case XK_s: *dy = 1; break;
	    case XK_Right:
	    case XK_KP_Right:
	    case XK_KP_6:
	    case XK_E:
	    case XK_e: *dx = 1; break;
	    case XK_Left:
	    case XK_KP_Left:
	    case XK_KP_4:
	    case XK_W:
	    case XK_w: *dx = -1; break;
	    case XK_KP_Home:
	    case XK_KP_7: *dy = -1; *dx = -1; break;
	    case XK_KP_Prior:
	    case XK_KP_9: *dy = -1; *dx = 1; break;
	    case XK_KP_End:
	    case XK_KP_1: *dy = 1; *dx = -1; break;
	    case XK_KP_Next:
	    case XK_KP_3: *dy = 1; *dx = 1; break;
	    case XK_Q:
	    case XK_q: game_over = 1; break;
	    }
	  continue;
	}
      else if (event->type == ButtonRelease)
	{
	  map_window->get_offset (&event->xbutton, dx, dy);
	  continue;
	}
    }
}


// yes_no_event_loop()
// Process X events until a question is answered
Boolean
yes_no_event_loop (void)
{
  XEvent *event;
  KeySym keysym;
  char buffer[4];

  while (!game_over)
    {
      event = adventure_window->next_event ();
      if (event->type == KeyPress)
	{
	  XLookupString (&event->xkey, buffer, sizeof (buffer), &keysym, NULL);
	  switch (keysym)
	    {
	    case XK_Y:
	    case XK_y:
	      return True;
	    case XK_N:
	    case XK_n:
	      return False;
	    case XK_Q:
	    case XK_q: game_over = 1; break;
	    }
	}
    }
  return False;
}


// number_event_loop()
// Process X events until a number is chosen
int
number_event_loop (const char *message, int min, int max)
{
  XEvent *event;
  KeySym keysym;
  char buffer[40], strnum[12];
  int answer = min, i = 0, ch;

  sprintf (buffer, message, answer);
  transient->print (TM_ADD, buffer);
  while (!game_over)
    {
      event = adventure_window->next_event ();
      if (event->type == KeyPress)
	{
	  XLookupString (&event->xkey, buffer, sizeof (buffer), &keysym, NULL);
	  switch (keysym)
	    {
	    case XK_0:
	    case XK_KP_0: ch = '0'; break;
	    case XK_1:
	    case XK_KP_1: ch = '1'; break;
	    case XK_2:
	    case XK_KP_2: ch = '2'; break;
	    case XK_3:
	    case XK_KP_3: ch = '3'; break;
	    case XK_4:
	    case XK_KP_4: ch = '4'; break;
	    case XK_5:
	    case XK_KP_5: ch = '5'; break;
	    case XK_6:
	    case XK_KP_6: ch = '6'; break;
	    case XK_7:
	    case XK_KP_7: ch = '7'; break;
	    case XK_8:
	    case XK_KP_8: ch = '8'; break;
	    case XK_9:
	    case XK_KP_9: ch = '9'; break;
	    case XK_plus:
	    case XK_KP_Add: ch = '+'; break;
	    case XK_minus:
	    case XK_hyphen:
	    case XK_KP_Subtract: ch = '-'; break;
	    case XK_BackSpace:
	    case XK_Delete: ch = '\b'; break;
	    case XK_Clear:
	    case XK_Escape: ch = '\e'; break;
	    case XK_Return:
	    case XK_KP_Equal:
	    case XK_KP_Enter: ch = '\n'; break;
	    case XK_Q:
	    case XK_q: game_over = 1;
	    default: ch = '\0';
	    }

	  if ((ch >= '0') && (ch <= '9'))
	    {
	      // Add another digit to the number string.
	      if (i < 10)
		{
		  strnum[i++] = ch;
		  strnum[i] = '\0';
		} else {
		  adventure_window->noise ();
		}
	      answer = strtoul (strnum, NULL, 10);
	      if (answer > max)
		answer = max;
	      if (answer < min)
		answer = min;
	    }
	  else if (ch == '\b')
	    {
	      if (i)
		strnum[--i] = '\0';
	      else
		adventure_window->noise ();
	      answer = strtoul (strnum, NULL, 10);
	      if (answer > max)
		answer = max;
	      if (answer < min)
		answer = min;
	    }
	  else if (ch == '\e')
	    {
	      i = 0;
	      strnum[i] = '\0';
	      answer = min;
	    }
	  else if (ch == '-')
	    {
	      answer--;
	      if (answer < min)
		answer = min;
	      sprintf (strnum, "%d", answer);
	      i = strlen (strnum);
	    }
	  else if (ch == '+')
	    {
	      answer++;
	      if (answer > max)
		answer = max;
	      sprintf (strnum, "%d", answer);
	      i = strlen (strnum);
	    }
	  else if (ch == '\n')
	    break;
	}
      sprintf (buffer, message, answer);
      transient->print (TM_REPLACE, buffer);
    }

  return answer;
}


// timed_event_loop()
// Process X events until a specified number of milliseconds has passed
void
timed_event_loop (int ms)
{
  adventure_window->event_pause (ms);
  return;
}


// update_status()
void
update_status (void)
{
  int weight;
  char msgbuf[128];

  weight = 0;
  if (game_data.Food > 0)
    weight += game_data.Food * 500;
  if (game_data.Gold > 0)
    weight += game_data.Gold * 3;
  if (game_data.Water > 0)
    weight += game_data.Water * 1500;
  adventure_window->clear_status ();
  sprintf (msgbuf, "HERO - Hit Points: %d, Gold: %d, Weight: %d.%d, Food: %d, Water: %d\n",
	  game_data.HitPoints, game_data.Gold, weight / 3000,
	   ((weight + 150) / 300) % 10, game_data.Food, game_data.Water);
  adventure_window->print (msgbuf);
  if (game_data.PartySize == 2)
    {
      sprintf (msgbuf, "COMPANION - Hit Points: %d\n",
	       game_data.HitPointsCompanion);
      adventure_window->print (msgbuf);
    }
  if (game_data.Interest && game_data.DebtToCompanion)
    {
      sprintf (msgbuf, "Money owed: $%d to Abdul, $%d to companion",
	       game_data.Interest, game_data.DebtToCompanion);
      adventure_window->print (msgbuf);
    }
  else if (game_data.Interest && !game_data.DebtToCompanion)
    {
      sprintf (msgbuf, "Money owed: $%d to Abdul", game_data.Interest);
      adventure_window->print (msgbuf);
    }
  else if (!game_data.Interest && game_data.DebtToCompanion)
    {
      sprintf (msgbuf, "Money owed: $%d to companion",
	       game_data.DebtToCompanion);
      adventure_window->print (msgbuf);
    }
}
